#Common RestApi methods
API_COMMON_DEVICE_OFF = "off"
API_COMMON_DEVICE_ON = "on"
API_COMMON_GET_STATE = "get_state/"
API_COMMON_SET_STATE = "set_state/"
API_COMMON_GET_DEVICE_INFO = "get_device_info/"

#Client connection timeout in seconds
API_CLIENT_CONNECTION_TIMEOUT = 5

#Common RestApi status values
API_RESPONSE_STATUS_OK = "ok"
API_RESPONSE_STATUS_FAIL = "false"
API_RESPONSE_STATUS_INVALID = "invalid_action_name"

#Request base params keys
REQUEST_CHANNEL_KEY = "channel"
REQUEST_STATE_KEY = "state"

#Logger lib name
LOGGER_NAME = "foxrestapiclient_log"


"""F&F Fox device responses implmentation."""

from foxrestapiclient.connection import _LOGGER

from .const import API_RESPONSE_STATUS_FAIL


class RestApiError:
    """RestApi error data holder."""

    def __init__(self, error):
        """Construct object."""
        self.error = error

class RestApiBaseResponse:
    """Base response from Fox any device."""

    def __init__(self, status: str, error: str = "", error_obj = None) -> None:
        """Construct object with status and optional error params."""
        self.status = status
        self.__has_errors = False
        if error != "" or error_obj is not None:
            self.__has_errors = True
        self.error = error
        self.__error_obj = error_obj

    def has_errors(self) -> bool:
        """Return hasErrors flag."""
        return self.__has_errors

    def get_error_obj(self):
        """Get error object."""
        return self.__error_obj

class RestApiDeviceStateResponse(RestApiBaseResponse):
    """F&F Fox device state response."""

    def __init__(self, channel_1_state: str = None, channel_2_state: str = None,
                overcurrent: str = None, status: str = API_RESPONSE_STATUS_FAIL,
                state: str = None, error_obj = None) -> None:
        """Construct object with provided parameters."""
        super().__init__(status, error_obj=error_obj)
        self.state = state
		#Some of Fox devices has two channels support
        self.channel_1_state = channel_1_state
        self.channel_2_state = channel_2_state
		#Some of Fox devices has overcurrent property, we can check if device can work.
        self.overcurrent = overcurrent

class RestApiDeviceInfoResponse(RestApiBaseResponse):
    """F&F Fox device info response."""

    def __init__(self, device_name: str = "", firmware: str = "", hw: str = "",
                updater: str = "", device_friendly_name: str = "",
                device_commercial_name: str = "", status: str = API_RESPONSE_STATUS_FAIL,
                device_channels_name: list = None, error_obj = None) -> None:
        """Construct object with provided parameters."""
        super().__init__(status, error_obj=error_obj)
        self.device_name = device_name
        self.firmware = firmware
        self.hardware = hw
        self.updater = updater
        self.device_friendly_name = device_friendly_name
        self.device_commercial_name = device_commercial_name
        self.device_channels_name = device_channels_name

class RestApiBrightnessResponse(RestApiBaseResponse):
    """F&F Fox device brightness response."""

    def __init__(self, channel_1_value: str = None,  channel_2_value: str = None, value: str = None,
                 status: str = API_RESPONSE_STATUS_FAIL, error_obj = None) -> None:
        """Construct object with provided parameters."""
        super().__init__(status, error_obj=error_obj)
        self.brightness  = -1
        self.channel_1_value = -1
        self.channel_2_value = -1

        try:
            if value is not None:
                self.brightness = int(value)
            if channel_1_value is not None:
                self.channel_1_value = int(channel_1_value)
            if channel_2_value is not None:
                self.channel_2_value = int(channel_2_value)
        except Exception as exception:
            _LOGGER.error("Error in parsing, %s", exception)



from . import fox_base_device
from . import fox_dim1s2_device
from . import fox_dimmable_device
from . import fox_led2s2_device
from . import fox_r1s1_device
from . import fox_r2s2_device
from . import fox_rgbw_device
from . import fox_str1s2_device


"""F&F Fox STR1S2 device implementation."""

from __future__ import annotations

import json

from foxrestapiclient.connection import _LOGGER
from foxrestapiclient.connection.const import (API_RESPONSE_STATUS_FAIL,
                                               API_RESPONSE_STATUS_OK)
from foxrestapiclient.connection.rest_api_client import RestApiClient
from foxrestapiclient.connection.rest_api_responses import RestApiBaseResponse

from .const import (API_STR1S2_GET_OPEN_LEVEL, API_STR1S2_GET_TILT_LEVEL,
                    API_STR1S2_SET_OPEN_LEVEL, API_STR1S2_SET_TILT_LEVEL)
from .fox_base_device import DeviceData, FoxBaseDevice


class FoxSTR1S2Device(FoxBaseDevice):
    """STR1S2 rollershutter device."""

    def __init__(self, device_data: DeviceData):
        """Initalize object."""
        super().__init__(device_data)
        #State reprsents cover open or close
        self._state = False
        self.__device_api_client = self.DeviceRestApiImplementer(self._rest_api_client)
        self._cover_position = 0
        self._tilt_position = 0

    class CoverOpenLevel(RestApiBaseResponse):
        """Cover open level data holder."""

        def __init__(self, level: str = None, status: str = API_RESPONSE_STATUS_FAIL) -> None:
            """Initialize object."""
            super().__init__(status)
            self.level = 0
            try:
                self.level = int(level)
            except:
                self.level = 0

    class DeviceRestApiImplementer:
        """Specific RestAPI methods definition used by device."""

        def __init__(self, rest_api_client: RestApiClient) -> None:
            """Initalize object."""
            self._rest_api_client = rest_api_client

        async def async_fetch_level(self, method: str) -> FoxSTR1S2Device.CoverOpenLevel:
            """Get open level api method."""
            device_response = await self._rest_api_client.async_make_api_call_get(method)
            if device_response is None:
                return FoxSTR1S2Device.CoverOpenLevel(status=API_RESPONSE_STATUS_FAIL)
            return FoxSTR1S2Device.CoverOpenLevel(**json.loads(device_response))

        async def async_get_open_level(self) -> FoxSTR1S2Device.CoverOpenLevel:
            """Get cover open level."""
            return await self.async_fetch_level(API_STR1S2_GET_OPEN_LEVEL)

        async def async_get_tilt_level(self) -> FoxSTR1S2Device.CoverOpenLevel:
            """Get tilt level."""
            return await self.async_fetch_level(API_STR1S2_GET_TILT_LEVEL)

        async def update_open_level(self, params, method)  -> RestApiBaseResponse:
            """Set open level api method."""
            device_response = await self._rest_api_client.async_make_api_call_get(method, params)
            if device_response is None:
                return RestApiBaseResponse(status=API_RESPONSE_STATUS_FAIL)
            return RestApiBaseResponse(**json.loads(device_response))

        async def async_set_open_level(self, params) -> RestApiBaseResponse:
            """Set open cover level."""
            return await self.update_open_level(params, API_STR1S2_SET_OPEN_LEVEL)

        async def async_set_tilt_open_level(self, params) -> RestApiBaseResponse:
            """Set tilt open level."""
            return await self.update_open_level(params, API_STR1S2_SET_TILT_LEVEL)

    def is_on(self, channel: int = None):
        """Return device state."""
        _LOGGER.warning("This device does not support is on funcionality.")

    def is_cover_opened(self):
        """Get is cover open.

        Open means position is greater than 0.
        """
        return self._cover_position > 0

    def is_cover_closed(self):
        """Get is cover closed."""
        return self._cover_position == 0

    def get_cover_position(self):
        """Get cover postion."""
        return self._cover_position

    def get_tilt_position(self):
        """Get tilt position."""
        return self._tilt_position

    async def __async_fetch_open_level(self, callback = None, is_cover = True):
        """Fetch cover open level."""
        if callback is None:
            return
        cover_open = await callback()
        if cover_open.status != API_RESPONSE_STATUS_OK:
            self._state = False
            return
        if is_cover:
            self._cover_position = cover_open.level
        else:
            self._tilt_position = cover_open.level

    async def async_fetch_cover_open_level(self):
        """Fetch cover open level."""
        await self.__async_fetch_open_level(self.__device_api_client.async_get_open_level)

    async def async_fetch_tilt_open_level(self):
        """Fetch tilt open level."""
        await self.__async_fetch_open_level(self.__device_api_client.async_get_tilt_level, False)

    async def __async_cover_set_level(self, level = 0, is_cover = True) -> bool:
        """Set cover level.

        Keyword arguments:
        level -- cover position must be in range <0,100>
        callback -- function to call

        Return: True if success, False otherwise.
        """
        if level < 0 or level > 100:
            _LOGGER.warning(
                "Level passed to __async_cover_set_level is out of range. Supported <0,100>")
            return False
        params = {
            "level": level
        }
        if is_cover is True:
            device_response = await self.__device_api_client.async_set_open_level(params)
        else:
            device_response = await self.__device_api_client.async_set_tilt_open_level(params)
        if device_response.status != API_RESPONSE_STATUS_OK:
            self._state = False
            return False
        self._state = True
        return True

    async def async_open_cover(self) -> bool:
        """Open cover.

        Return: True if success, False otherwise
        """
        return await self.__async_cover_set_level(100)

    async def async_close_cover(self) -> bool:
        """Close cover.

        Return: True if success, False otherwise
        """
        return await self.__async_cover_set_level(0)

    async def async_set_cover_position(self, position) -> bool:
        """Set cover position.

        Keyword arguments:
        position -- position to set up from 0 to 100%

        Return: True if success, False otherwise."""
        return await self.__async_cover_set_level(position)

    async def async_set_tilt_positon(self, position) -> bool:
        """Set tilt position.

        Keyword arguments:
        position -- position to set up from 0 to 100%

        Return: True if success, False otherwise.
        """
        return await self.__async_cover_set_level(position, False)

    async def async_set_device_state(self, is_on, channel = None):
        """Overriden method from base device."""
        _LOGGER.warning("This device does not support set_device_state funcionality.")

    async def async_fetch_channel_state(self, channel: int = None) -> bool | list:
        """Overriden method from base device."""
        _LOGGER.warning("This device does not support fetch_channel_state funcionality.")

    async def async_update_channel_state(self, state: bool, channel: int = None) -> bool:
        """Overriden method from base device."""
        _LOGGER.warning("This device does not support update_channel_state funcionality.")

    async def async_fetch_update(self):
        """Fetch all available data from device."""
        await self.async_fetch_cover_open_level()
        await self.async_fetch_tilt_open_level()



"""F&F Fox R2S2 device implementation."""

from foxrestapiclient.connection import _LOGGER

from .fox_base_device import DeviceData, FoxBaseDevice


class FoxR2S2Device(FoxBaseDevice):
    """F&F Fox R2S2 device implementation. Two relays and two switches."""

    def __init__(self, device_data: DeviceData):
        """Initalize object."""
        super().__init__(device_data)
        #This device has two channels
        self.channels = [1, 2]
        self.channel_one_state = False
        self.channel_two_state = False

    def is_on(self, channel: int = None):
        """Return device is on status.

        Return F&F Fox R2S2 device status. Check if it is on by given channel.

        Keyword arguments:
        channel -- required parameter to obtain status. Can be 1 or 2.
        Return:
        channel state, on or off.
        """
        if channel < 0 or channel > 2:
            _LOGGER.warning("Passed channel in is_on() has wrong value. Only 1 or 2 are accepted.")
            return False
        return self.channel_one_state if channel == 1 else self.channel_two_state

    def __set_channels_to_off(self):
        """Private method to reseting channels state."""
        self.channel_one_state = False
        self.channel_two_state = False

    async def async_fetch_update(self):
        """Abstract method implementation. Fetch all required data from device."""
        states = await self.async_fetch_channel_state()
        if states is not None and isinstance(states, list):
            self.channel_one_state, self.channel_two_state = states
        else:
            self.__set_channels_to_off()



"""F&F Fox LED2S2 device implementation."""

from foxrestapiclient.connection import _LOGGER

from .fox_base_device import DeviceData
from .fox_dimmable_device import FoxDimmableDevice


class FoxLED2S2Device(FoxDimmableDevice):
    """LED2S2 Device implementation."""

    def __init__(self, device_data: DeviceData):
        """Initialze object."""
        super().__init__(device_data)
        #This device has two channels
        self.channels = [1, 2]
        self.channel_one_state = False
        self.channel_two_state = False
        #Store brighntess value for each channel
        self.channel_one_brightness = 0
        self.channel_two_brightness = 0

    def is_on(self, channel: int = None) -> bool:
        """Return device state to determine that is on or off.

        Keyword arguments:
        channel -- channel number to check that is on or off. Must be in ranfe <1,2>
        """
        if channel < 0 or channel > 2:
            _LOGGER.warning("Channel provided to is_on() is out of range.")
            return False
        return self.channel_one_state if channel == 1 else self.channel_two_state

    async def async_fetch_channel_one_state(self) -> bool:
        """Fetch device state on channel one."""
        self.channel_one_state = await self.async_fetch_channel_state(self.channels[0])
        return self.channel_one_state

    async def async_fetch_channel_two_state(self) -> bool:
        """Fetch device state on channel two."""
        self.channel_one_state = await self.async_fetch_channel_state(self.channels[1])
        return self.channel_two_state

    async def async_update_channel_one_state(self, state: bool):
        """Update device channel one state by given value."""
        self.channel_one_state = await self.async_update_channel_state(state, self.channels[0])
        return self.channel_one_state

    async def async_update_channel_two_state(self, state: bool):
        """Update device channel one state by given value."""
        self.channel_two_state = await self.async_update_channel_state(state, self.channels[1])
        return self.channel_two_state

    async def async_fetch_channel_one_brightness(self) -> int:
        """Fetch channel one brightness value."""
        channel_brightness = await self.async_fetch_channel_brightness(self.channels[0])
        if isinstance(channel_brightness, list):
            self.channel_one_brightness = channel_brightness[0]
        else:
            self.channel_one_brightness = 0
        return self.channel_one_brightness

    async def async_fetch_channel_two_brightness(self) -> int:
        """Fetch channel two brightness value."""
        channel_brightness = await self.async_fetch_channel_brightness(self.channels[1])
        if isinstance(channel_brightness, list):
            self.channel_two_brightness = channel_brightness[0]
        else:
            self.channel_two_brightness = 0
        return self.channel_two_brightness

    async def async_update_channel_one_brightness(self):
        """Update channel one brightness."""
        self.channel_one_brightness = await self.async_update_channel_brightness(self.channels[0])
        return self.channel_one_brightness

    async def async_update_channel_two_brightness(self):
        """Update channel two brightness."""
        self.channel_two_brightness = await self.async_update_channel_brightness(self.channels[1])
        return self.channel_two_brightness

    def __reset_channels_state(self):
        """Reset channels state."""
        self.channel_one_state = False
        self.channel_two_state = False

    def __reset_channels_brightness(self):
        """Reset channels brightness."""
        self.channel_one_brightness = 0
        self.channel_two_brightness = 0

    async def async_fetch_update(self):
        """Fetch all available data from device."""
        states = await self.async_fetch_channel_state()
        if not isinstance(states, list):
            self.__reset_channels_state()
        else:
            self.channel_one_state, self.channel_two_state = states
        brightness = await self.async_fetch_channel_brightness()
        if not isinstance(brightness, list):
            self.__reset_channels_brightness()
        elif isinstance(brightness, list) and len(brightness) < 2:
            self.__reset_channels_brightness()
        else:
            self.channel_one_brightness, self.channel_two_brightness = brightness



"""F&F Fox base device implementation."""
from __future__ import annotations

import abc

from aiohttp.client_exceptions import ClientConnectionError

from foxrestapiclient.connection.rest_api_client import RestApiClient
from foxrestapiclient.connection.rest_api_responses import \
    RestApiDeviceInfoResponse

from .const import (API_RESPONSE_STATUS_FAIL, API_RESPONSE_STATUS_INVALID, API_RESPONSE_STATUS_OK,
                    DEVICE_OFF, DEVICE_ON, DEVICE_PLATFORM, DEVICES,
                    MANUFACTURER_NAME)


class DeviceData:
    """DeviceData holder. Used for simple object creation."""

    def __init__(self, name: str, host: str, api_key: str, mac_addr: str, dev_type: int,
            channels: list = None):
        """Init all required values."""
        self.name = name
        self.host = host
        self.api_key = api_key
        self.mac_addr = mac_addr
        self.dev_type = dev_type
        self.channels = channels

class UnsupportedDevice(Exception):
    """Custom exception for unsupported device."""


class FoxBaseDevice:
    """F&F Fox base device class."""

    __metaclass__ = abc.ABCMeta
    def __init__(self, device_data: DeviceData):
        """Construct object with data passed. Warning! can be exception raised."""
        self.dev_type = device_data.dev_type
        self.name = device_data.name
        self.mac_addr = device_data.mac_addr
        self.device_info_data: RestApiDeviceInfoResponse = None
        self.is_available = False
        self._rest_api_client = RestApiClient(device_data.host, device_data.api_key)
        self.__init_device_platform(device_data.dev_type)

    def __init_device_platform(self, dev_type: int):
        """Initialize device platform.

        Keyword arguments:
        type -- device type, determine platfom by this value.

        Warning!
        If type not exsist in DEVICE_PLATFORM exception UnsupportedDevice will be raised.
        """
        try:
            self.device_platform = DEVICE_PLATFORM[dev_type]
            self._rest_api_client.register_response_error_hook(self.__track_available)
        except:
            raise UnsupportedDevice("Not supported device. Check type param.")

    def equals(self, fox_base_device: FoxBaseDevice) -> bool:
        """Compare object and return that are equal.

        Comparation works on unique identifier.

        Keyword arguments:
        foxBaseDevice -- device base object to compare to

        Return:
        true if devices are equal.
        """
        if fox_base_device is None:
            return False
        if self.mac_addr == fox_base_device.mac_addr:
            return True
        return False

    def __track_available(self, error):
        """Track device availability."""
        if isinstance(error, ClientConnectionError):
            self.is_available = False
        if error is None:
            self.is_available = True

    def get_channel_name(self, channel: int) -> str:
        """Return channel name or name."""
        if (self.device_info_data is None or (channel < 1 and channel > 2)):
            return self.name
        if len(self.device_info_data.device_channels_name) > 1:
            return self.device_info_data.device_channels_name[channel-1]
        return self.name

    def get_device_info(self) -> str:
        """Get device info JSON string.

        This method is similar to HomeAssistant get_device_info()
        Can be easy adapted to HA system.

        Return:
        JSON string contains device info.
        """
        name = self.name
        sw_version = "0.0.0"
        if self.device_info_data is not None:
            name = self.device_info_data.device_friendly_name
            sw_version = self.device_info_data.firmware

        return {
            "identifiers": {
                # Serial numbers are unique identifiers within a specific domain
                (self.device_platform, self.mac_addr)
            },
            "name": name,
            "manufacturer": MANUFACTURER_NAME,
            "model": DEVICES[self.dev_type],
            "sw_version": sw_version
        }

    async def async_fetch_channel_state(self, channel: int = None) -> bool | list:
        """Fetch device state.

        Fetch F&F Fox device state by given channel. If None provided
        device get_state generic method will be used. This method can
        return list bool values or single bool value.

        Keyword arguments:
        channel -- optional, fetch state by given channel

        Return:
        list[bool] - if no channel provided and device has more than one channel
                     index 0 idicates to channel 1, index 1 to channel 2
        bool - device has only one channel, true or false.
        """
        device_response = await self._rest_api_client.async_api_get_device_state(channel)
        if device_response.status in (API_RESPONSE_STATUS_FAIL, API_RESPONSE_STATUS_INVALID):
            self.is_available = False
            return False
        self.is_available = True
        if device_response.state == DEVICE_ON:
            return True
        if device_response.state == DEVICE_OFF:
            return False
        values = []
        if device_response.channel_1_state == DEVICE_ON:
            values.append(True)
        else:
            values.append(False)
        if device_response.channel_2_state == DEVICE_ON:
            values.append(True)
        else:
            values.append(False)
        return values if values != [] else [False, False]

    async def async_update_channel_state(self, state: bool, channel: int = None) -> bool:
        """Update device state.

        Update F&F Fox device state. Turn it on or off.

        Keyword arguments:
        state -- value to set on device, true or false, where true means turn on.
        channel -- optional, set state by given channel.

        Return:
        true or false depends on device response.
        """
        device_response = await self._rest_api_client.async_api_set_device_state(state, channel)
        if device_response.status in (API_RESPONSE_STATUS_FAIL, API_RESPONSE_STATUS_INVALID):
            return False
        return device_response.status == API_RESPONSE_STATUS_OK

    async def async_fetch_device_info(self):
        """Fetch device info.

        Fetch F&F Fox device info. Name, firmware version etc. see RestApiDeviceInfoResponse.
        Warning! this method will override device name provided in constructor. Device name defined
        in Fox mobile application has priority.
        """
        device_response: RestApiDeviceInfoResponse = (
            await self._rest_api_client.async_api_get_device_info()
        )
        if device_response.status in (API_RESPONSE_STATUS_FAIL, API_RESPONSE_STATUS_INVALID):
            return False
        self.device_info_data = device_response
        #Overwrite device name from user app config
        if device_response.device_name is not None:
            self.name = device_response.device_name
        return True

    async def async_fetch_device_available_data(self):
        """Fetch all available data from device."""
        await self.async_fetch_device_info()
        await self.async_fetch_update()

    @abc.abstractmethod
    def is_on(self, channel: int = None):
        """Return device state to determine that is on or off."""

    @abc.abstractmethod
    async def async_fetch_update(self):
        """Abstract method - fetch data from device."""



#Common RestApi status values
API_RESPONSE_STATUS_OK = "ok"
API_RESPONSE_STATUS_FAIL = "false"
API_RESPONSE_STATUS_INVALID = "invalid_action_name"

#Common RestApi methods
API_METHOD_COMMON_GET_STATE = "get_state/"
API_METHOD_COMMON_SET_STATE = "set_state/"
API_METHOD_COMMON_GET_DEVICE_INFO = "get_device_info/"

#Devices RestApi methods
API_R1S1_GET_AC_PARAMETERS = "get_current_energy/"
API_R1S1_GET_TOTAL_ENERGY_DATA = "get_total_energy/"

API_DIMMABLE_SET_BRIGHTNESS = "set_brightness/"
API_DIMMABLE_GET_BRIGHTNESS = "get_brightness/"

API_RGBW_SET_COLOR_HSV = "set_color_hsv/"
API_RGBW_GET_COLOR_HSV = "get_color_hsv/"

API_STR1S2_GET_OPEN_LEVEL = "get_open_level/"
API_STR1S2_SET_OPEN_LEVEL = "set_open_level/"
API_STR1S2_GET_TILT_LEVEL = "get_open_louvers_level/"
API_STR1S2_SET_TILT_LEVEL = "set_open_louvers_level/"

#Device model
DEVICE_MODEL_DIM1S2 = "DIM1S2"
DEVICE_MODEL_GATE = "GATE"
DEVICE_MODEL_LED2S2 = "LED2S2"
DEVICE_MODEL_POWER = "POWER"
DEVICE_MODEL_R1S1 = "R1S1"
DEVICE_MODEL_R2S2 = "R2S2"
DEVICE_MODEL_RGBW = "RGBW"
DEVICE_MODEL_STR1S2= "STR1S2"
DEVICE_MODEL_SUN1 = "SUN1"

#Device state values
DEVICE_OFF = "off"
DEVICE_ON = "on"

#Device type
DEVICE_TYPE_DIM1S2 = 9
DEVICE_TYPE_GATE = 1
DEVICE_TYPE_LED2S2 = 7
DEVICE_TYPE_POWER = 0 #T0D0
DEVICE_TYPE_R1S1 = 8
DEVICE_TYPE_R2S2 = 4
DEVICE_TYPE_RGBW = 6
DEVICE_TYPE_STR1S2 = 2
DEVICE_TYPE_SUN1 = 10

#Device model dictionary, key: device type
DEVICES = {
    DEVICE_TYPE_GATE: DEVICE_MODEL_GATE,
    DEVICE_TYPE_STR1S2: DEVICE_MODEL_STR1S2,
    DEVICE_TYPE_R2S2: DEVICE_MODEL_R2S2,
    DEVICE_TYPE_RGBW: DEVICE_MODEL_RGBW,
    DEVICE_TYPE_LED2S2: DEVICE_MODEL_LED2S2,
    DEVICE_TYPE_R1S1: DEVICE_MODEL_R1S1,
    DEVICE_TYPE_DIM1S2: DEVICE_MODEL_DIM1S2,
    DEVICE_TYPE_SUN1: DEVICE_MODEL_SUN1
}

#Manufacturer info
MANUFACTURER_NAME = "F&F Filipowski Sp.j"

#F&F Fox device supporting platform
SUPPORTED_PLATFORM_LIGHT = "light"
SUPPORTED_PLATFORM_SWITCH = "switch"
SUPPORTED_PLATFORM_SENSOR = "sensor"
SUPPORTED_PLATFORM_COVER = "cover"
SUPPORTED_PLATFORM_GATE = "gate"

#Device platform map
DEVICE_PLATFORM = {
    DEVICE_TYPE_DIM1S2: SUPPORTED_PLATFORM_LIGHT,
    DEVICE_TYPE_LED2S2: SUPPORTED_PLATFORM_LIGHT,
    DEVICE_TYPE_RGBW: SUPPORTED_PLATFORM_LIGHT,
    DEVICE_TYPE_R1S1: SUPPORTED_PLATFORM_SWITCH,
    DEVICE_TYPE_R2S2: SUPPORTED_PLATFORM_SWITCH,
    DEVICE_TYPE_STR1S2: SUPPORTED_PLATFORM_COVER
}

#Service discovery values
DEVICE_DISCOVERY_RESPONSE_HEADER = 'F&F-WiFi-device-discovery-response:1'
DEVICE_DISCOVERY_REQUEST_HEADER = 'F&F-WiFi-device-discovery-request:1'
MIN_DATA_SIZE_TO_PARSE = 45


"""Service discovery for F&F Fox devices."""
import asyncio
from typing import Tuple

from foxrestapiclient.connection import _LOGGER

from .const import (DEVICE_DISCOVERY_REQUEST_HEADER,
                    DEVICE_DISCOVERY_RESPONSE_HEADER, DEVICES,
                    MIN_DATA_SIZE_TO_PARSE)
from .fox_base_device import DeviceData


class DeviceDiscoverProtocol(asyncio.DatagramProtocol):
    """Device disover protocol used in asyncio service discovery implementation."""

    def __init__(self, datagram_parser_callback = None) -> None:
        """Construct object."""
        super().__init__()
        self._datagram_parser_callback = datagram_parser_callback

    def datagram_received(self, data: bytes, addr: Tuple[str, int]) -> None:
        """Datagram reveived.

        Invoked when UDP datagram is available to read.

        Keyword arguments:
        data -- byte array response from device
        addr -- typle with sender ip address and port.

        """
        if self._datagram_parser_callback:
            self._datagram_parser_callback(data, addr)

    def error_received(self, exc):
        """Error received in UDP data."""
        _LOGGER.error("Exception thrown in UDP data receiving: %s", exc)

class FoxServiceDiscovery:
    """Discover F&F Fox devices in local network over UDP broadcast."""

    def __init__(self) -> None:
        """Construct object."""
        self._loop = asyncio.get_running_loop()
        self._discovered_devices = []

    def get_discovered_devices(self):
        """Get dicovered devices."""
        return self._discovered_devices

    def __check_exsist_device(self, device: DeviceData):
        for dev in self._discovered_devices:
            if dev.mac_addr == device.mac_addr:
                return True
        return False

    async def async_discover_devices(self, default_tries = 5) -> list:
        """Async discovering devices.

        Send request to all devices in local network to discover them.

        Keyword arguments:
        default_tries -- default probes to discover Fox devices. This
            value extends discovering time by following formula:
            default_tries * 4 s = discovering time in seconds.

        Return: discovered devices list.
        """
        transport, protocol = await self._loop.create_datagram_endpoint(
            lambda: DeviceDiscoverProtocol(self.parse_received_datagram),
            local_addr=('0.0.0.0', 1395),
            reuse_port=True,
            allow_broadcast=True
        )
        #Clear discovered devices list.
        self._discovered_devices = []
        try:
            for i in range(default_tries):
                #Send message to UDP broadcast
                transport.sendto(
                    DEVICE_DISCOVERY_REQUEST_HEADER.encode(),
                    ('255.255.255.255', 1918)
                )
                await asyncio.sleep(4)
        finally:
            transport.close()
        return self._discovered_devices

    def parse_received_datagram(self, data: bytes, addr):
        """Parse UDP message."""
        if len(data) < MIN_DATA_SIZE_TO_PARSE:
            _LOGGER.warring("Received data size is not enough to parsing it.")
        if len(data) < len(DEVICE_DISCOVERY_RESPONSE_HEADER.encode()):
            _LOGGER.error("Parsing UDP response error. Cannot discover F&F Fox device.")
            return
        if data[0:36].decode() != DEVICE_DISCOVERY_RESPONSE_HEADER:
            _LOGGER.warning("Response not indicate to F&F Fox device.")
            return
        device_type = int.from_bytes(data[42:44], "little")
        device_unique_id = data[36:42].hex()
        try:
            discovered_device = DeviceData(
                DEVICES[device_type],
                addr[0],
                "000",
                device_unique_id,
                device_type
            )
            _LOGGER.info("Received datagram from %s. Successfuly parsed.", addr)
            if self.__check_exsist_device(discovered_device) is False:
                self._discovered_devices.append(discovered_device)
        except KeyError:
            _LOGGER.error("Unsupported! F&F Fox device not implmeneted yet.")



"""Fox dimmable device implementation."""

import json

from foxrestapiclient.connection import _LOGGER
from foxrestapiclient.connection.const import (API_RESPONSE_STATUS_FAIL,
                                               API_RESPONSE_STATUS_INVALID,
                                               REQUEST_CHANNEL_KEY)
from foxrestapiclient.connection.rest_api_client import RestApiClient
from foxrestapiclient.connection.rest_api_responses import (
    RestApiBaseResponse, RestApiBrightnessResponse)

from .const import API_DIMMABLE_GET_BRIGHTNESS, API_DIMMABLE_SET_BRIGHTNESS
from .fox_base_device import DeviceData, FoxBaseDevice


class FoxDimmableDevice(FoxBaseDevice):
    """Fox Dimmable device implementation. Base class for any device with dimmable feature."""

    def __init__(self, device_data: DeviceData):
        """Initialize object."""
        super().__init__(device_data)
        #Extened RestApi methods, specific for device
        self.__device_api_client = self.DeviceRestApiImplementer(self._rest_api_client)

    class DeviceRestApiImplementer:
        """Inner class with specific RestApi methods definition used by device."""

        def __init__(self, rest_api_client: RestApiClient) -> None:
            """Initialize object."""
            self._rest_api_client = rest_api_client

        async def async_get_brightness_value(self, params) -> RestApiBrightnessResponse:
            """Get brightness value by given channel.

            Keyword arguments:
            params -- params dictionary, should contain channel number
            Return:
            RestApiBrightnessResponse
            """
            device_response = await self._rest_api_client.async_make_api_call_get(
                API_DIMMABLE_GET_BRIGHTNESS, params)
            if device_response is None:
                return RestApiBrightnessResponse(status=API_RESPONSE_STATUS_FAIL)
            return RestApiBrightnessResponse(**json.loads(device_response))

        async def async_set_brighntess_value(self, params) -> RestApiBaseResponse:
            """Set brightness value.

            Keyword arguments:
            params -- params dictionary, should contain channel number
            Return:
            RestApiBaseResponse
            """
            device_response = await self._rest_api_client.async_make_api_call_get(
                API_DIMMABLE_SET_BRIGHTNESS, params)
            if device_response is None:
                return RestApiBaseResponse(API_RESPONSE_STATUS_FAIL)
            return RestApiBaseResponse(**json.loads(device_response))

    async def async_fetch_channel_brightness(self, channel: int = 0) -> list:
        """Fetch device brightness value by given channel.

        If channel number not provided fetch brightness from all channels.

        Keyword arguments:
        channel -- channel number
        Return:
        list(int) -- readed values.
        """
        params = None
        if channel != 0:
            params = {
                REQUEST_CHANNEL_KEY: str(channel)
            }
        device_response = await self.__device_api_client.async_get_brightness_value(params)
        if device_response.status in (API_RESPONSE_STATUS_FAIL, API_RESPONSE_STATUS_INVALID):
            return [0,0]
        values = []
        if device_response.brightness != -1:
            values.append(device_response.brightness)
        if device_response.channel_1_value != -1:
            values.append(device_response.channel_1_value)
        if device_response.channel_2_value != -1:
            values.append(device_response.channel_2_value)
        return values

    async def async_update_channel_brightness(self, brightness: int, channel: int = None) -> bool:
        """Set brightness value on device.

        Keyword arguments:
        brightness -- value from range <0-255>
        channel -- value from range <1-2> or None.
        """
        if brightness < 0 or brightness > 255:
            _LOGGER.warning(
                "Brightness passed to sync_update_channel_brightness() is out of range.")
            return False
        if channel is not None and (channel < 0 or channel > 2):
            _LOGGER.warning(
                "Channel passed to sync_update_channel_brightness() is out of range.")
            return False
        params = {
            "value": brightness
        }
        if channel is not None:
            params.update({REQUEST_CHANNEL_KEY: channel})
        device_response = await self.__device_api_client.async_set_brighntess_value(params)
        if device_response.status in (API_RESPONSE_STATUS_FAIL, API_RESPONSE_STATUS_INVALID):
            _LOGGER.error("Setting brightness value in async_update_channel_brightness() failed.")
        return True




"""F&F Fox DIM1S2 device implementation."""

from .fox_base_device import DeviceData
from .fox_dimmable_device import FoxDimmableDevice


class FoxDIM1S2Device(FoxDimmableDevice):
    """DIM1S2 device one channel dimmer 230V."""

    def __init__(self, device_data: DeviceData):
        """Initalize object."""
        super().__init__(device_data)
        self.brightness = 0
        self.state = False

    def is_on(self, channel: int = None):
        """Return device state idicates to turn on or off."""
        return self.state

    async def async_fetch_update(self):
        """Fetch all available data from device."""
        self.state = await self.async_fetch_channel_state()
        channel_brightness = await self.async_fetch_channel_brightness()
        if isinstance(channel_brightness, list):
            self.brightness = channel_brightness[0]
        else:
            self.brightness = 0
